from Point import calc_distance
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
import urllib.request
from osgeo import gdal, ogr, osr
import os
import time
import datetime
from dataset import Dataset
from math import ceil as round_up
from geopy import distance

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
DISTRIBUTION_LOCATION = "\\\\igskrdcnpssmbfs.ngtoc-n.usgs.gov\\cldsr\\Elevation\\PRD\\Output\\"


class TNMAPIException(Exception):
    def __init__(self, response, message='TNM API Bad Response: '):
        self.response_code = str(response)
        super(TNMAPIException, self).__init__(message + self.response_code)


def set_output_feature(feature, pt, out_layer):
    # Fill in field info
    # Iterate through values from input feature and fill fields
    DEFAULT_VAL = -9999
    for key in pt.field_dict:
        feature.SetField(key, pt.field_dict[key])
    if pt.original_z:
        feature.SetField("Orig_Z", pt.original_z)
    if pt.original_x:
        feature.SetField("Orig_X", pt.original_x)
    if pt.original_y:
        feature.SetField("Orig_Y", pt.original_y)
    feature.SetField("One_3rd_Z", pt.one_third_z) if pt.one_third_z else feature.SetField("One_3rd_Z", DEFAULT_VAL)
    feature.SetField("Five_M_Z", pt.five_meter_z) if pt.five_meter_z else feature.SetField("Five_M_Z", DEFAULT_VAL)
    feature.SetField("One_M_Z", pt.one_meter_z) if pt.one_meter_z else feature.SetField("One_M_Z", DEFAULT_VAL)
    feature.SetField("LPC_Z_Valu", pt.lpc_z) if pt.lpc_z else feature.SetField("LPC_Z_Valu", DEFAULT_VAL)
    feature.SetField("Final_z", pt.final_z)
    feature.SetField("Final_Res", pt.final_res_name)

    if pt.final_z and pt.original_z:
        feature.SetField("ZDiff", pt.final_z - pt.original_z)
    # Set the geometry
    gdal.PushErrorHandler('CPLQuietErrorHandler')
    if pt.lpc_x and pt.lpc_y:
        wkt = "POINT(%.10f %.10f)" % (pt.lpc_x, pt.lpc_y)
        feature.SetField("Final_X", pt.lpc_x)
        feature.SetField("Final_Y", pt.lpc_y)
    elif pt.center_x and pt.center_y:
        wkt = "POINT(%f %f)" % (pt.center_x, pt.center_y)
        feature.SetField("Final_X", pt.map_corrected_summit_x)
        feature.SetField("Final_Y", pt.map_corrected_summit_y)
        # Calculate the distance between the two points in meters
        # feature.SetField("Distance", calc_distance(pt.original_x, pt.original_y, pt.map_corrected_summit_x, pt.map_corrected_summit_y))

        # Note: the distance.geodesic() function is not set up to account for altitude and in situations where the
        # difference in altitude is great (I think > 100m) then the accuracy would be affected
        dist = distance.geodesic((pt.original_y, pt.original_x), (pt.map_corrected_summit_y, pt.map_corrected_summit_x))
        feature.SetField("Distance", dist.m)
    else:
        wkt = "POINT(%f %f)" % (pt.map_x, pt.map_y)
    point_geom = ogr.CreateGeometryFromWkt(wkt)
    print("Original X and Y: ", pt.original_x, pt.original_y)
    feature.SetGeometry(point_geom)
    out_layer.CreateFeature(feature)


def sort_funct(e):
    return datetime.datetime.strptime(e["publicationDate"], "%Y-%m-%d")


def tnm_api_download_raster(bbox, out_loc, data_type, lock, reping=None, debug=None, bad_list=None):
    pointX, pointY, _, _ = bbox.split(',')

    url = "https://tnmaccess.nationalmap.gov/api/v1/products"
    if not reping:
        with lock:
            print("Pinging TNM for " + data_type + " " + bbox)
    parameters = {"bbox": bbox, "datasets": data_type}
    num_bad_response = 0
    while True:
        try:
            response = requests.get(url, params=parameters, verify=False)
            if not response.ok:
                num_bad_response += 1
                if num_bad_response > 20:
                    raise TNMAPIException(response.status_code)
            else:
                if debug:
                    print("URL: " + response.request.url)
                try:
                    data = response.json()
                except Exception as bad_response:
                    with lock:
                        print(bad_response)
                    return None

                if data['total'] == 0:
                    if not reping:
                        with lock:
                            print("No data found from: " + response.request.url)
                    return None
                valid_index = []
                item_index = None
                item_date = None
                download_url = None
                dist_loc = None

                # Compare returned files
                for i in range(0, len(data['items'])):
                    if not data['items'][i]['downloadURL'].endswith('.tif'):
                        continue
                    if float(data['items'][i]['boundingBox']['minX']) < float(pointX) < \
                            float(data['items'][i]['boundingBox']['maxX']) and \
                            float(data['items'][i]['boundingBox']['minY']) < float(pointY) < \
                            float(data['items'][i]['boundingBox']['maxY']):
                        tmp_item_date = datetime.datetime.strptime(data['items'][i]["publicationDate"], "%Y-%m-%d").date()
                        if bad_list and data['items'][i]['title'].replace(" ", "_").replace('1/3', '13') not in bad_list:
                            if item_index is None or item_date < tmp_item_date:
                                item_index = i
                                item_date = tmp_item_date
                if item_index is None:
                    return None
                download_url = data['items'][item_index]['downloadURL']
                title = data['items'][item_index]['title'].replace(" ", "_")
                if '1/3' in title:
                    title = title.replace('1/3', '13')
                if data_type == "Lidar Point Cloud (LPC)":
                    downloaded_raster = os.path.join(out_loc, title + ".laz")
                else:
                    downloaded_raster = os.path.join(out_loc, title + ".tif")

                if data_type == "Digital Elevation Model (DEM) 1 meter,Alaska IFSAR 5 meter DEM":
                    if 'one_meter' in title:
                        data_type = "Digital Elevation Model (DEM) 1 meter"
                    else:
                        data_type = "Alaska IFSAR 5 meter DEM"
                if not os.path.isfile(downloaded_raster):
                    with lock:
                        print("Downloading dataset: " + downloaded_raster)
                    _, http_message = urllib.request.urlretrieve(download_url, downloaded_raster)
                else:
                    if not reping:
                        with lock:
                            print("Dataset: " + downloaded_raster + " found in output location.")
                    if bad_list and downloaded_raster in bad_list:
                        with lock:
                            print("Dataset in bad location")
                        return None
                    # If file exists check it's last modification time. If it has been modified in the last 75 seconds wait
                    time_since_mod = get_last_modification_diff(downloaded_raster)
                    while time_since_mod < datetime.timedelta(seconds=75):
                        time.sleep(5)
                        time_since_mod = get_last_modification_diff(downloaded_raster)
                d = dataset(data_type, title, downloaded_raster)
                d.bad_location = bad_list if bad_list is not None else []
                return d
        except:
            time.sleep(1)
        else:
            if response.status_code == 400:
                time.sleep(30)
                # continue
        break


def build_1_3_tile_name(pointX, pointY):
    tile_name = ""
    nameX = float(pointX)
    nameY = float(pointY)
    if nameY > 0:
        tile_name += "n" + str(round_up(abs(float(nameY))))
    else:
        tile_name += "s" + str(round_up(abs(float(nameY))))
    if nameX > 0:
        tile_name += "e" + str(round_up(abs(float(nameX))))
    else:
        tile_name += "w" + str(round_up(abs(float(nameX))))
    return tile_name


# returns difference between time last modified and current time
def get_last_modification_diff(downloaded_raster):
    ti_m = os.path.getmtime(downloaded_raster)
    return datetime.timedelta(seconds=(time.time() - ti_m))


def truncate(num, n):
    integer = int(num * (10 ** n)) / (10 ** n)
    return float(integer)
