from math import cos, sin, asin, sqrt, radians
from osgeo import osr, ogr

# https://en.wikipedia.org/wiki/Earth_radius#Mean_radius
EARTH_MEAN_RADIUS = 6371008.8
EARTH_MEAN_DIAMETER = 2 * EARTH_MEAN_RADIUS


# Class to store information for each summit point
class Point:
    map_x = None
    map_y = None
    feature_id = None
    name = None
    feature = None
    state = None
    county = None
    corrected_summit_x = None
    corrected_summit_y = None
    map_corrected_summit_x = None
    map_corrected_summit_y = None
    corrected_value = None
    center_x = None
    center_y = None
    lpc_x = None
    lpc_y = None
    lpc_z = None
    original_x = None
    original_y = None
    original_z = None
    one_third_z = None
    five_meter_z = None
    one_meter_z = None
    final_res = None
    final_z = None
    final_res_name = None
    read_error_NED = None
    read_error_5M = None

    def __init__(self, x, y, field_dict, multi=None):
        self.datasets = []
        self.map_x = x
        self.map_y = y
        self.field_dict = field_dict
        self.final_z = None
        self.final_res_name = None
        self.name = 'No "FEATURE_NA" or "gaz_name" field found'
        if 'FEATURE_NA' in self.field_dict:
            self.name = self.field_dict['FEATURE_NA']
        elif 'gaz_name' in self.field_dict:
            self.name = self.field_dict['gaz_name']
        self.multi = multi

    def get_next_resolution(self):
        d = self.get_most_recent_data()
        if d.type == "National Elevation Dataset (NED) 1/3 arc-second":
            return "Digital Elevation Model (DEM) 1 meter,Alaska IFSAR 5 meter DEM"
        elif d.type == "Alaska IFSAR 5 meter DEM":
            return None
        elif d.type == "Digital Elevation Model (DEM) 1 meter":
            return "Lidar Point Cloud (LPC)"
        return None

    def get_most_recent_data(self):
        if self.datasets:
            return self.datasets[-1]
        else:
            return None

    def get_pdal_bbox(self):
        one_meter_dataset = self.get_most_recent_data()
        x_max = self.center_x + .5 * one_meter_dataset.cell_size_x
        x_min = self.center_x - .5 * one_meter_dataset.cell_size_x
        y_max = self.center_y + .5 * one_meter_dataset.cell_size_y
        y_min = self.center_y - .5 * one_meter_dataset.cell_size_y
        return "([{}, {}], [{}, {}])".format(truncate(x_min, 5), truncate(x_max, 5),
                                             truncate(y_min, 5), truncate(y_max, 5))

    # TODO: look at getting lat/lon without transform to WGS84
    def get_bbox(self, in_srs):
        # Set the bbox for tnm api
        if self.center_x and self.center_y:
            # Transform from Web Mercator to WGS84
            if str(in_srs.GetAttrValue("Unit")).upper() != "DEGREE":
                targetSR = osr.SpatialReference()
                targetSR.ImportFromEPSG(4326)  # WGS84
                targetSR.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
                trans_point = ogr.Geometry(ogr.wkbPoint)
                trans_point.AddPoint(self.center_x, self.center_y)
                trans_point.AssignSpatialReference(in_srs)
                trans_point.TransformTo(targetSR)
                bbox_xmin = trans_point.GetX()
                bbox_xmax = trans_point.GetX()
                bbox_ymin = trans_point.GetY()
                bbox_ymax = trans_point.GetY()
            else:
                # Build bbox
                # print("actual point: ([{}, {}])".format(self.center_x, self.center_y))
                bbox_xmin = self.map_x
                bbox_xmax = self.map_x
                bbox_ymin = self.map_y
                bbox_ymax = self.map_y
            return "{},{},{},{}".format(bbox_xmin, bbox_ymin, bbox_xmax, bbox_ymax)
        else:
            # Transform from Web Mercator to WGS84
            if str(in_srs.GetAttrValue("Unit")).upper() != "DEGREE":
                targetSR = osr.SpatialReference()
                targetSR.ImportFromEPSG(4326)  # WGS84
                targetSR.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
                # coordTrans = osr.CoordinateTransformation(in_srs, targetSR)   DOESN'T GET USED
                trans_point = ogr.Geometry(ogr.wkbPoint)
                trans_point.AddPoint(self.map_x, self.map_y)
                trans_point.AssignSpatialReference(in_srs)
                trans_point.TransformTo(targetSR)
                print("actual point: ([{}, {}])".format(trans_point.GetX(), trans_point.GetY()))
                bbox_xmin = trans_point.GetX()
                bbox_xmax = trans_point.GetX()
                bbox_ymin = trans_point.GetY()
                bbox_ymax = trans_point.GetY()
            else:
                print("actual point: ([{}, {}])".format(self.map_x, self.map_y))
                bbox_xmin = self.map_x
                bbox_xmax = self.map_x
                bbox_ymin = self.map_y
                bbox_ymax = self.map_y
            return "{},{},{},{}".format(bbox_xmin, bbox_ymin, bbox_xmax, bbox_ymax)

    def set_info(self, xmin=None, ymax=None, pt_srs=None, raster_srs=None, res_name=None):
        # Get most recent dataset
        recent_data = self.get_most_recent_data()
        if not recent_data.type == "Lidar Point Cloud (LPC)":
            #  Put the point in the middle of the cell
            self.center_x = (xmin + (self.corrected_summit_x * recent_data.cell_size_x)) + (0.5 * recent_data.cell_size_x)
            self.center_y = (ymax - (self.corrected_summit_y * recent_data.cell_size_y)) - (0.5 * recent_data.cell_size_y)
            # put point coord back in pt_srs
            pt_srs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
            trans_point = ogr.Geometry(ogr.wkbPoint)

            trans_point.AddPoint(self.center_x, self.center_y)
            trans_point.AssignSpatialReference(raster_srs)
            trans_point.TransformTo(pt_srs)
            self.center_x, self.center_y = trans_point.GetX(), trans_point.GetY()

            if recent_data.type == "National Elevation Dataset (NED) 1/3 arc-second":
                # Set the one third z value
                self.one_third_z = self.corrected_value
            elif recent_data.type == "Digital Elevation Model (DEM) 1 meter":
                # Set the one meter z value
                self.one_meter_z = self.corrected_value
            elif recent_data.type == "Alaska IFSAR 5 meter DEM":
                # Set the five meter z value
                self.five_meter_z = self.corrected_value
        else:
            # Set the lpc z value
            self.lpc_z = self.corrected_value
        # Set the final z and res for the point
        if (not self.final_z) or self.corrected_value > self.final_z:
            self.final_z = self.corrected_value
            self.final_res_name = res_name

        # set final_z
        # self.final_z = self.corrected_value


# Returns distance in meters based on the haversine formula
# Doesn't seem to account for
def calc_distance(lat1, lon1, lat2, lon2):
    """
    Calculate the great circle distance between two points
    on the earth (specified in decimal degrees)
    # The haversines (sin(...)^2) are wrong. Need to use theta/2 = (d/r)/2
    """
    # convert decimal degrees to radians
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])
    # haversine formula
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
    c = 2 * asin(sqrt(a))
    km = 6371 * c
    return km * 1000


def truncate(num, n):
    integer = int(num * (10**n))/(10**n)
    return float(integer)


def haversine_distance(point1, point2):
    """
        Calculating haversine distance between two points
        (see https://en.wikipedia.org/wiki/Haversine_formula,
            https://www.math.ksu.edu/~dbski/writings/haversine.pdf)
        Is numerically better-conditioned for small distances
    """
    lon1, lat1 = (radians(coord) for coord in point1[:2])
    lon2, lat2 = (radians(coord) for coord in point2[:2])
    dlat = (lat2 - lat1)
    dlon = (lon2 - lon1)
    a = (
        sin(dlat * 0.5)**2 +
        cos(lat1) * cos(lat2) * sin(dlon * 0.5)**2
    )

    return EARTH_MEAN_DIAMETER * asin(sqrt(a))
